/*
    \brief Deklarationen für die Funktionen der Geräteverbindungen und deren direktem Zugriff.

    \author Bernd Ottmann
    \date   21.02.2003
*/
#pragma once
#ifndef DATAFOX_CONNECTION_H
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define DATAFOX_CONNECTION_H
#endif

#include "df_defines.h"

/*!
    \brief Initialisiert eine serielle Verbindung für die Kommunikation mit einem AEIII+ über RS232/RS485.

    Die Funktion ruft intern DFCComOpenSerial() mit den Standardparametern
    \a baudRate = 38400 und \a timeout = 800 auf.

    \ingroup connection

    \export{_DFCComInit@4, DFCComInit}

    \attention Diese Funktion sollten Sie nur für die Kommunikation mit AEIII+ Geräten verwenden, bei allen anderen Geräten
    sollten Sie DFCComOpenIV() zum initalisieren einer Verbindung nutzen.
    Die Kommunikation startet bei Verwendung mit anderen Geräten verzögert, da vorher das zugrundeliegende
    Kommunikationsprotokoll ermittelt werden muss.
    Sie ist aufgrund des verwendeten Timeout's für Modemverbindungen nicht einzusetzen.

    \param[in] connectionId Zu initalisierende [Verbindungs-Id](\ref dfc_connection_id). Diese Nummer wird intern zu
    COM[connectionId] oder /etc/ttyS[connectionId - 1] formatiert.
    Eine Angabe von 4 führt Beispielsweise zu einem Aufruf von DFCComOpenSerial( 4, "COM4", 38400, 800 )
    bei Windows oder DFCComOpenIV( 4, "/etc/ttyS3", 38400, 800 ) bei Linux.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComInit( int connectionId );

/*!
    \brief Initalisiert eine serielle Verbindung für die Kommunikation mit einem AEIII+ für RS232 / RS485.

    \ingroup connection

    \export{_DFCComOpenSerial@16, DFCComOpenSerial}

    \note Unter Windows kann die Schnittstelle als "\\.\COM1" angeben werden. Ab Nummern > 9 wird dieses sogar
    notwendig.
    In C-Style Zeichenfolge würden Sie dieses wegen der nötigen Maskierung als "\\\\.\\COM10" schreiben.

    \param[in] connectionId Zu initialisierende [Verbindungs-Id](\ref dfc_connection_id).
    \param[in] commPort Name der Schnittstelle "COM1", "COM2", "/etc/ttyS0" usw..
    \param[in] baudRate Baudrate mit der die Schnittstelle initalisiert wird. Standardwert 38400.
    \param[in] timeout Zu verwendender [Schnittstellentimeout](\ref dfc_read_timeout) in Millisekunden. Standardwert 1000.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenSerial( int connectionId, const char *commPort, int baudRate, int timeout );

/*!
    \brief Initalisiert eine TCP/IP Verbindung.

    Die Funktion ruft intern DFCComOpenSocket() mit den Standardparametern
    \a portNumber = 8000 und \a timeout = 3000 auf.

    \ingroup connection

    \export{_DFCComOpen@8, DFCComOpen}

    \warning Nach einem Verbindungsabriss wird keine automatische Neuverbindung durchgeführt.\n
    Bitte beachten Sie hierzu die Fehlerbehandlung und unter Exemplarischer Programmaufbau beschriebene Umsetzungslösung.

    \param[in] connectionId Zu initalisierende [Verbindungs-Id](\ref dfc_connection_id).
    \param[in] hostAddress Hostname oder IP - Adresse in Punktnotation.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpen( int connectionId, const char *hostAddress );

/*!
    \brief Initalisiert eine TCP/IP Verbindung.

    \ingroup connection

    \export{_DFCComOpenSocket@16, DFCComOpenSocket}

    \warning Nach einem Verbindungsabriss wird keine automatische Neuverbindung durchgeführt.\n
    Bitte beachten Sie hierzu die Fehlerbehandlung und unter Exemplarischer Programmaufbau beschriebene Umsetzungslösung.

    \param[in] connectionId Zu initalisierende [Verbindungs-Id](\ref dfc_connection_id).
    \param[in] hostAddress Hostname oder IP - Adresse in Punktnotation.
    \param[in] portNumber Portnummer. Standardwert 8000.
    \param[in] timeout Zu verwendender [Schnittstellentimeout](\ref dfc_read_timeout) in Millisekunden. Standardwert 3000.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenSocket( int connectionId, const char *hostAddress, int portNumber, int timeout );

/*!
    \brief Initalisiert die Kommunikationsverbindung.

    Initalisiert eine Verbindung zu ggf. mehreren Geräten. Nach erfolgreicher Ausführung kann mit der hier initalisierten
    Verbindungs-Id die anderen Funktionen der Bibliothek aufgerufen werden.
    Mittels DFCComClose() kann eine bestehende Verbindung geschlossen werden.

    \ingroup connection

    \export{_DFCComOpenIV@24, DFCComOpenIV}

    \note Diese Funktion führt noch keine Kommunikation mit einem Gerät durch. Die erfolgreiche Ausführung stellt somit
    nur sicher, dass die serielle Schnittstelle ordnungsgemäß initialisiert werden konnte oder eine entsprechende
    TCP/IP Verbindung hergestellt wurde.

    \warning Bei TCP/IP Verbindungen wird nach einem Verbindungsabriss keine automatische Neuverbindung durchgeführt.\n
    Bitte beachten Sie hierzu die Fehlerbehandlung und unter Exemplarischer Programmaufbau beschriebene Umsetzungslösung.


    \param[in] connectionId Zu initalisierende [Verbindungs-Id](\ref dfc_connection_id).
    \param[in] deviceAddress Die von der Bibliothek zu verwendende Gerätenummer. Standardwert ist 0.
    Hierbei handelt es sich nicht um die Gerätenummer des später anzusprechenden Gerätes. Diese
    Angabe hier wird als Absenderadresse in den Protokollpacketen verwendet.
    \param[in] connectionType Der zu verwendende Schnittstellentyp.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Bedeutung</th></tr>
        <tr><td>1</td><td>RS232 direkt oder über Umsetzter (RS232 <-> RS485)</td></tr>
        <tr><td>2</td><td>Funk über den Umsetzer (RS232 <-> BIM2) ohne eigenen Kontroller.</td></tr>
        <tr><td>3</td><td>TCP / IP - Verbindung.</td></tr>
    </table>
    \param[in] connectionString Je nach gewähltem Schnittstellentyp:
    <table class="paramvalues">
        <tr><th>connectionType</th><th>Bedeutung</th></tr>
        <tr><td>1, 2</td><td>Zeiger auf Zeichenkette für die Schnittstelle. "COM1" .. "COM4".</td></tr>
        <tr><td>3</td><td>Zeiger auf Zeichenkette mit IP - Adresse oder Hostname.</td></tr>
    </table>
    \param[in] connectionValue Je nach gewähltem Schnittstellentyp:
    <table class="paramvalues">
        <tr><th>connectionType</th><th>Bedeutung</th></tr>
        <tr><td>1, 2</td><td>Übertragungsrate der Schnittstelle. (Standard ist: 38400)</td></tr>
        <tr><td>3</td><td>Portnummer des TCP / IP Anschlusses.</td></tr>
    </table>
    \param[in] comunicationReadTimeout Je nach gewähltem Schnittstellentyp:
    <table class="paramvalues">
        <tr><th>connectionType</th><th>Bedeutung</th></tr>
        <tr><td>1, 2</td><td>Timeout für Lesen von Schnittstelle (Standard ist: 800)</td></tr>
        <tr><td>3</td><td>Timeout für Lesen von Socket (Standard ist: 3000)</td></tr>
    </table>

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.

    \par Beispiel
    \snippet examples.cpp DFCComOpenIV
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenIV( int connectionId, int deviceAddress, int connectionType, 
                                                    const char *connectionString, int connectionValue, int comunicationReadTimeout );

/*!
    \brief Schließen einer geöffneten Verbindung.

    Je nach dem zugrundeliegenden Typ der Verbindungs-Id, wird entweder die serielle Schnittstelle oder
    die TCP/IP Verbindung geschlossen.

    \ingroup connection

    \export{_DFCComClose@4, DFCComClose}

    \param[in] connectionId Zu schließende Verbindungs-Id.

    \snippet examples.cpp DFCComClose
*/
DF_API_EXTERN_DECORATE void DF_API_CALL DFCComClose( int connectionId );

#if (DF_API_VERSION >= 0x04030000)

/*!
    \brief Ruft den aktuellen Zustand einer Verbindungs-Id ab.

    \ingroup connection

    \export{_DFCIsChannelOpen@4, DFCIsChannelOpen}

    \param[in] connectionId Zu verwendende Verbindungs-Id.

    \retval -1 Verbindungsobjekt nicht verfügbar, Verbindungs-Id wurde noch nicht verwendet.
    \retval 0 Verbingungs-Id hält eine geschlossene Verbindung.
    \retval 1 Verbindungs-Id hält eine geöffnete Verbindung.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCIsChannelOpen( int connectionId );

#endif

/*!
    \brief Sendet Daten ohne Verwendung des Datafox-Protokolls.

    Diese Funktion ist für die Initalisierung von Zwischengeräten wie Modems gedacht (per AT-Kommandos).
    Es gibt auch die Möglichkeit die Hauptkommunikation der Datafox Geräte in einen transparenten Modus zu schalten
    und über diese Funktion Daten an ein dort angeschlossenes Drittgerät wie CNC-Maschinen zu senden.

    \ingroup connection

    \export{_DFCWrite@20, DFCWrite}

    \warning Die Funktion sollte nur für Sonderfälle wie Modeminitialisierung verwendet werden!
    Durch unsachgemäßen Gebrauch, können Störungen in der Kommunikation auftreten.

    \param[in] connectionId Zu verwendende[Verbindungs-Id.
    \param[in] buffer Puffer der zu sendenden Datenbytes.
    \param[in] toWrite Anzahl der Datenbytes.
    \param[out] written Anzahl der geschriebenen Datenbytes.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCWrite( int connectionId, const char *buffer, int toWrite, int *written, int *errorNumber );

/*!
    \brief Empfängt Daten ohne Verwendung des Datafox-Protokolls.

    Diese Funktion ist für die Initalisierung von Zwischengeräten wie Modems gedacht (per AT-Kommandos).
    Es gibt auch die Möglichkeit die Hauptkommunikation der Datafox Geräte in einen transparenten Modus zu schalten
    und über diese Funktion Daten von einem dort angeschlossenes Drittgerät wie CNC-Maschinen zu lesen.

    \ingroup connection

    \export{_DFCRead@20, DFCRead}

    \warning Die Funktion sollte nur für Sonderfälle wie Modeminitialisierung verwendet werden!
    Durch unsachgemäßen Gebrauch, können Störungen in der Kommunikation auftreten.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] buffer Puffer für die zu empfangenen Datenbyte.
    \param[in] toRead Anzahl der maximal in buffer zu speichernden Datenbyte.
    \param[out] read Anzahl der in buffer gespeicherten Datenbyte.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCRead( int connectionId, char *buffer, int toRead, int *read, int *errorNumber );

// ########################################################
// #############* PRIVATE FUNKTIONEN* #################
// ########################################################

// Funktionen sind nicht dokumentiert.
// Sie werden von AESetup oder Talk verwendet.
// Unterliegen evntl. Aenderungen.

#if defined( USE_PRIVATE )

//    \export{_DFCComUseHandle@8, DFCComUseHandle}
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComUseHandle( int connectionId, HANDLE commHandle );

//    \export{_DFCComReleaseHandle@4, DFCComReleaseHandle}
DF_API_EXTERN_DECORATE int DF_API_CALL DFCComReleaseHandle( int connectionId );

//    \export{_DFCGetComPort@4, DFCGetComPort}
DF_API_EXTERN_DECORATE HANDLE DF_API_CALL DFCGetComPort( int connectionId );

//    \export{_DFCSetComPort@20, DFCSetComPort}
DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetComPort( int connectionId, HANDLE handle, int isSocket, int value, int timeout );

#endif // USE_PRIVATE

#endif
