#include "common/software_info.h"
#include <boost/format.hpp>
#include <memory.h>

#define sizeof_StructSoftwareInfo_1    25 // Strukturgroesse der Version 1 ist: 25Byte
#define sizeof_StructSoftwareInfo_2    30 // Strukturgroesse der Version 2 ist: 30Byte
#define sizeof_StructSoftwareInfo_3    32 // Strukturgroesse der Version 3 ist: 32Byte
#define sizeof_StructSoftwareInfo_4    36 // Strukturgroesse der Version 4 ist: 36Byte
#define sizeof_StructSoftwareInfo_5    42 // Strukturgroesse der Version 5 ist: 42Byte
#define sizeof_StructSoftwareInfo_6    50 // Strukturgroesse der Version 6 ist: 50Byte

SoftwareInfo::SoftwareInfo()
{
    clear();
}

SoftwareInfo::~SoftwareInfo()
{
}

void SoftwareInfo::clear()
{
    // Sturkturdaten löschen.
    memset( &_softwareInfo, 0, sizeof(StructSoftwareInfo) );
}

void SoftwareInfo::setStructSoftwareInfo( const StructSoftwareInfo &info )
{
    _softwareInfo = check( info );
}

StructSoftwareInfo SoftwareInfo::structSoftwareInfo() const
{
    return _softwareInfo;
}

StructSoftwareInfo SoftwareInfo::check( StructSoftwareInfo swInfo )
{
    if ( swInfo.size > sizeof(StructSoftwareInfo) )
    {
        // Zu aktuell fuer diese Softwareversion
        memset( &swInfo, 0, sizeof(StructSoftwareInfo) );
        return swInfo;
    }

    // Anpassungen fuer neuere Versionen durchfuehren
    if ( swInfo.size < sizeof_StructSoftwareInfo_2 )
    {
        // Auf Version 2 anpassen
        swInfo.comserver         = COMSERVER_ENABLE; // Comserver aktiv
        swInfo.transponder_write = (swInfo.transponder > 0) ? TP_WRITE_ENABLE : TP_WRITE_NONE; // Transponder schreiben (koennen eigentlich nicht alle)
        swInfo.http              = (swInfo.modem > 0) ? HTTP_ENABLE : HTTP_NONE; // HTTP-Protokoll, wenn Modem vorhanden.
        swInfo.language          = LANGUAGE_ENABLE; // Sprachdatei wird unterstützt. Bedeutet nicht, dass in alten Geraetedateien diese da sein müssen.
        swInfo.biokey_update     = (swInfo.fingerprint > 0) ? BIOKEY_UPDATE_ENABLE : BIOKEY_UPDATE_NONE; // Biokey updaten, wenn Fingerprint vorhanden.
    } 

    // Anpassungen fuer neuere Versionen durchfuehren
    if ( swInfo.size < sizeof_StructSoftwareInfo_3 )
    {
        // Auf Version 3 anpassen
        swInfo.setup_options      = SETUP_OPTION_ALL; // Alle Optionen des Setups werden unterstützt
    } 

    // Anpassungen fuer neuere Versionen durchfuehren
    if ( swInfo.size < sizeof_StructSoftwareInfo_4 )
    {
        // Auf Version 4 anpassen
        swInfo.touch   = TOUCH_NONE;
        swInfo.sensor  = SENSOR_NONE;
        swInfo.battery = BATTERY_NO;
        swInfo.charge  = CHARGE_NO;
    } 

    // Anpassungen fuer neuere Versionen durchfuehren
    if ( swInfo.size < sizeof_StructSoftwareInfo_5 )
    {
        // Auf Version 5 anpassen
        swInfo.log = LOG_NONE;
        swInfo.payable_options = PAYABLE_OPTION_NONE;
        swInfo.transponderExtended = swInfo.transponder;
    } 

    // Anpassungen fuer neuere Versionen durchfuehren
    if ( swInfo.size < sizeof_StructSoftwareInfo_6 )
    {
        // Auf Version 6 anpassen
        swInfo.deviceExtended = swInfo.device;
        swInfo.boardExtended = swInfo.board;
    } 

    return swInfo;
}

std::string SoftwareInfo::toString()
{
    std::string res;
    if ( toString( _softwareInfo, res ) )
        return res;

    return std::string();
}

int SoftwareInfo::toString( const StructSoftwareInfo &swInfo, std::string &info )
{
    std::string temp;
    unsigned int bit;
    int res = 1;
    bool isMasterIV = false;

    // SofwareInfo verfuegbar?
    if ( !swInfo.size )
        return 0;

    temp.clear();

    info.append( "Firmwareinformation\r\n" );
    info.append( "------------------------------------------\r\n" );

    info.append( "\r\nBasis\r\n----------\r\n" );

    info.append( "Geraete: " );
    // Moegliche Geraetetypen
    for ( bit = 1; ; bit <<= 1 )
    {
        if ( bit & swInfo.deviceExtended )
        {
            if ( temp.size() )
                temp.append( ", " );

            switch ( bit )
            {
            case DEVICE_PZE:                temp.append( "PZE-MasterIV" ); isMasterIV = true; break;
            case DEVICE_ZK:                 temp.append( "ZK-MasterIV" );   isMasterIV = true; break;
            case DEVICE_BDE:                temp.append( "BDE-MasterIV" );  isMasterIV = true; break;
            case DEVICE_MOBIL:              temp.append( "Mobil-MasterIV" );  isMasterIV = true; break;
            case DEVICE_MDEBOX:             temp.append( "MDE-BoxIV" );  isMasterIV = true; break;
            case DEVICE_AE:                 temp.append( "AE-MasterIV" );  isMasterIV = true; break;
            case DEVICE_TIMEBOY:            temp.append( "TimeboyIV" );  isMasterIV = true; break;
            case DEVICE_EXKLUSIV:           temp.append( "Exklusiv-Line IV" );  isMasterIV = true; break;
            case DEVICE_FLEX:               temp.append( "Flex-MasterIV" );  isMasterIV = true; break;
            case DEVICE_TIMEBOY_MOBIL_PZE:  temp.append( "TimeboyIV Mobil-PZE" );  isMasterIV = true; break; // Diese Geräte sind alles MasterIV und ggf. EVO-Line mit Boardversion >= 4.0
            case DEVICE_TERMINAL_2_8:       temp.append( "EVO 2.8" ); isMasterIV = false; break;
            case DEVICE_TERMINAL_4_3:       temp.append( "EVO 4.3" ); isMasterIV = false; break;
            case DEVICE_IPC_EXTENSION_VARIO_5_7: temp.append( "IPC Embedded-System" ); isMasterIV = false; break;
            case DEVICE_IPC_EXTENSION_VARIO_10:  temp.append( "IPC Embedded-System" ); isMasterIV = false; break;
            case DEVICE_IPC_EXTENSION_Q7:        temp.append( "IPC Embedded-System" ); isMasterIV = false; break;
            case DEVICE_IO_BOX:             temp.append( "IO-Box V4" ); isMasterIV = false; break;
            case DEVICE_ZK_BOX:             temp.append( "ZK-Box V4" ); isMasterIV = false; break;
            case DEVICE_MOBIL_BOX:          temp.append( "Mobil-Box V4" ); isMasterIV = false; break;
            case DEVICE_TIMEBOY_TBDOCK:     temp.append( "Dockingstation V2" ); isMasterIV = false; break;
            case DEVICE_MOBIL_TBDOCK:       temp.append( "Fahrzeugdatenlogger V2" ); isMasterIV = false; break;
            case DEVICE_ZK_KNOTEN:          temp.append( "ZK-Knoten" ); isMasterIV = false; break;
            case DEVICE_EVO_3_5_UNIVERSAL:  temp.append( "EVO 3.5 Universal" ); isMasterIV = false; break;
            case DEVICE_SAMS_OCL_READER:    temp.append( "SAMS-OCL-Reader" ); isMasterIV = false; break;
            case DEVICE_TERMINAL_3_5:       temp.append( "EVO 3.5" ); isMasterIV = false; break;
            case DEVICE_ZK_READER_PLUS:     temp.append( "ZK-Leser Plus" ); isMasterIV = false; break;
            case DEVICE_ZK_ZENTRALE:        temp.append( "ZK-Zentrale" ); isMasterIV = false; break;
            default:                        temp.append( "Undefiniert" ); break;
            }
        }

        if ( bit == 0x80000000 )
            break;;
    }
    temp.append( "\r\n" );
    info.append( temp );
    temp.clear();

    info.append( "Platinenversionen: " );
    // Boardtyp pruefen
    for ( bit = 1; ; bit <<= 1 )
    {
        if ( bit & swInfo.boardExtended )
        {
            if ( temp.size() )
                temp.append( ", " );

            switch ( bit )
            {
            case BOARD_10: temp.append( "1.0" ); break;
            case BOARD_11: temp.append( "1.1" ); break;
            case BOARD_12: temp.append( "1.2" ); break;
            case BOARD_13: temp.append( "1.3" ); break;
            case BOARD_14: temp.append( "1.4" ); break;
            case BOARD_15: temp.append( "1.5" );  isMasterIV = false; break;
            case BOARD_16: temp.append( "1.6" );  isMasterIV = false; break;
            case BOARD_17: temp.append( "1.7" );  isMasterIV = false; break;
            case BOARD_18: temp.append( "1.8" );  isMasterIV = false; break;
            case BOARD_20: temp.append( "2.0" ); break;
            case BOARD_21: temp.append( "2.1" ); break;
            case BOARD_22: temp.append( "2.2" ); break;
            case BOARD_23: temp.append( "2.3" ); break;
            case BOARD_24: temp.append( "2.4" ); break;
            case BOARD_25: temp.append( "2.5" ); break;
            case BOARD_26: temp.append( "2.6" ); break;
            case BOARD_27: temp.append( "2.7" );  isMasterIV = false; break;
            case BOARD_30: temp.append( "3.0" ); break;
            case BOARD_31: temp.append( "3.1" ); break;
            case BOARD_34: temp.append( "3.4" ); break;
            case BOARD_40: temp.append( "4.0" );  isMasterIV = false; break; // Diese Boardversionen können bei PZE-MasterIV etc. verbaut sein, 
            case BOARD_41: temp.append( "4.1" );  isMasterIV = false; break; // dann handelt es sich aber nicht mehr um einen MasterIV im eigentlichen Sinne.
            case BOARD_42: temp.append( "4.2" );  isMasterIV = false; break;
            case BOARD_43: temp.append( "4.3" );  isMasterIV = false; break;
            case BOARD_44: temp.append( "4.4" );  isMasterIV = false; break;
            case BOARD_45: temp.append( "4.5" );  isMasterIV = false; break;
            case BOARD_46: temp.append( "4.6" );  isMasterIV = false; break;
			case BOARD_47: temp.append( "4.7" );  isMasterIV = false; break;
            case BOARD_48: temp.append( "4.8" );  isMasterIV = false; break;
            default:       temp.append( "Undefiniert" ); break;
            }
        }

        if ( bit == 0x80000000 )
            break;
    }
    temp.append( "\r\n" );
    info.append( temp );
    temp.clear();

    // Ab den Geraetetypen Terminal 2.8 werden die weiteren Pruefungen unterlassen.
    // Die noch "freien" Bits der Geraetetypen fallen auch unter diese Restriktion.
    if ( (swInfo.deviceExtended & 0xFFFFFC00) != 0 )
        return res;


    /////////////////////////////////////////////
    // Hauptkommunikationen
    // Wird nur bei MasterIV angezeigt.
    if ( isMasterIV )
    {
        info.append( "\r\nHauptkommunikationen\r\n------------------------------\r\n" );

        // Funk pruefen
        info.append( "Funk: " );
        if ( swInfo.radio )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.radio )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case RADIO_BLUETOOTH: temp.append( "Bluetooth" ); break;
                    case RADIO_ZIGBEE:    temp.append( "Zigbee" ); break;
                    default:              temp.append( "Undefiniert" ); break;
                    }

                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // ModemMCxx pruefen
        info.append( "Mobilfunk: " );
        if ( swInfo.modem )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.modem )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case MODEM_MC39: temp.append( "Siemens MC39" ); break;
                    case MODEM_MC55: temp.append( "Siemens MC55" ); break;
                    case MODEM_MODEM_RS9110_N_11_24: temp.append( "WLAN Redpine RS9110-N-11-24" ); break;
                    default:         temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // RS485 pruefen
        info.append( boost::str( boost::format("RS485: %1%\r\n") % (swInfo.rs485 ? "Ja" : "Nein") ) );

        // TCPIP pruefen
        info.append( "TCPIP: " );
        if ( swInfo.tcpip )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.tcpip )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case TCPIP_XPORT:     temp.append( "LAN" ); break;
                    case TCPIP_MATCHPORT: temp.append( "WLAN" ); break;
                    default:              temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // USB pruefen
        info.append( boost::str( boost::format("USB: %1%\r\n") % (swInfo.usb ? "Ja" : "Nein") ) );
    }

    /////////////////////////////////////////////
    // Zusatzoptionen
    info.append( "\r\nWaehlbare Optionen der Firmware\r\n------------------------------\r\n" );

    // RS485 (intern) ermitteln
    info.append( "Zutrittskontrolle: " );
    if ( swInfo.access )
    {
        for ( bit = 1; ; bit <<= 1 )
        {
            if ( bit & swInfo.access )
            {
                if ( temp.size() )
                    temp.append( ", " );

                switch ( bit )
                {
                case ACCESS_GIS:     temp.append( "TS-Serie" ); break;
                case ACCESS_PHG:     temp.append( "PHG-Serie" ); break;
                case ACCESS_PROTECT: temp.append( "Protect-Serie" ); break;
                case ACCESS_IDTRONIC: temp.append( "ID Tronic UHF-Serie" ); break;
                default:             temp.append( "Undefiniert" ); break;
                }
            }

            if ( bit == 0x80 )
                break;
        }
    }
    else
    {
        temp.append( "Nein" );
    }
    temp.append( "\r\n" );
    info.append( temp );
    temp.clear();

    // IOModule
    info.append( boost::str( boost::format("Anbindung IO-Modul: %1%\r\n") % (swInfo.iomodule ? "Ja" : "Nein") ) );

    // Timboyanbindung
    info.append( boost::str( boost::format("Anbindung TimeboyIII: %1%\r\n") % (swInfo.docking ? "Ja" : "Nein") ) );

    // Unterstuetzung COM-Server
    info.append( boost::str( boost::format("COM-Servermodus: %1%\r\n") % (swInfo.comserver ? "Ja" : "Nein") ) );

    // Unterstuetzung HTTP-Protokoll
    info.append( boost::str( boost::format("HTTP-Protokoll: %1%\r\n") % (swInfo.http ? "Ja" : "Nein") ) );

    // Transponder ueber das Transpondermenue schreiben
    info.append( boost::str( boost::format("Transponder ueber Transpondermenue schreiben: %1%\r\n") % (swInfo.transponder_write ? "Ja" : "Nein") ) );

    // Textdaten
    info.append( boost::str( boost::format("Textdaten benötigt: %1%\r\n") % (swInfo.language ? "Ja" : "Nein") ) );

    // BiokeyUpdate
    info.append( boost::str( boost::format("Biokeymodulupdate möglich: %1%\r\n") % (swInfo.biokey_update ? "Ja" : "Nein") ) );

    // Setupoptionen (intern) ermitteln
    info.append( boost::str( boost::format("Setupoptionen (0x%04X): ") % swInfo.setup_options ) );
    if ( swInfo.setup_options )
    {
        for ( bit = 1; ; bit <<= 1 ) 
        {
            // Nach maximal bekannter Option abbrechen
            if ( !(bit & SETUP_OPTION_MASK) )
                break;

            if ( bit & swInfo.setup_options )
            {
                if ( temp.size() )
                    temp.append( ", " );

                switch ( bit )
                {
                case SETUP_OPTION_IOCONTROL:     temp.append( "Signalverarbeitung" ); break;
                default:             temp.append( "Undefiniert" ); break;
                }
            }

            if ( bit == 0x8000 )
                break;
        }
    }
    else
    {
        temp.append( "Keine" );
    }
    temp.append( "\r\n" );
    info.append( temp );
    temp.clear();

    // Erweiterte Optionen ermitteln
    info.append( boost::str( boost::format("Erweiterte Optionen (0x%02X): ") % static_cast<unsigned short>( swInfo.payable_options ) ) );
    if ( swInfo.payable_options )
    {
        for ( bit = 1; ; bit <<= 1 ) 
        {
            // Nach maximal bekannter Option abbrechen
            if ( !(bit & PAYABLE_OPTION_MASK) )
                break;

            if ( bit & swInfo.payable_options )
            {
                if ( temp.size() )
                    temp.append( ", " );

                switch ( bit )
                {
                case PAYABLE_OPTION_DATA_ON_CARD: temp.append( "Data on Card" ); break;
                default:             temp.append( "Undefiniert" ); break;
                }
            }

            if ( bit == 0x80 )
                break;
        }
    }
    else
    {
        temp.append( "Keine" );
    }
    temp.append( "\r\n" );
    info.append( temp );
    temp.clear();

    /////////////////////////////////////////////
    // Zusatzbaugruppen
    // Wird nur bei MasterIV angezeigt.
    if ( isMasterIV )
    {
        info.append( "\r\nZusatzbaugruppen\r\n------------------------------\r\n" );

        // Display pruefen
        info.append( "Anzeige: " );
        if ( swInfo.display )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.display )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case DISPLAY_LEHNER_SWAP: temp.append( "4 Zoll Monochrome" ); break;
                    case DISPLAY_LEHNER:      temp.append( "6 Zoll Monochrome" ); break;
                    case DISPLAY_NEUMUELLER:  temp.append( "1,9 Zoll Color" ); break;
                    case DISPLAY_TIMEBOY_2ZOLL:  temp.append( "2 Zoll Color" ); break;
                    default:                  temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // Touch pruefen
        info.append( "Touch: " );
        if ( swInfo.touch )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.touch )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case TOUCH_SMALL: temp.append( "4 Zoll" ); break;
                    case TOUCH_BIG:   temp.append( "6 Zoll" ); break;
                    default:          temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // Sensor pruefen
        info.append( "Sensor: " );
        if ( swInfo.sensor )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.sensor )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case SENSOR_VIBRATION: temp.append( "Vibration" ); break;
                    case SENSOR_SPEEDUP:   temp.append( "Beschleunigung" ); break;
                    default:               temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // DigitalOut pruefen
        info.append( "Ausgaenge Digital: " );
        if ( swInfo.digital_out )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.digital_out )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit ) // Digitale Ausgaenge
                    {
                    case 0x01: temp.append( "dO1" ); break;
                    case 0x02: temp.append( "dO2" ); break;
                    case 0x04: temp.append( "dO3" ); break;
                    case 0x08: temp.append( "dO4" ); break;
                    default:   temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // AnalogIn pruefen
        info.append( "Eingaenge Analog: " );
        if ( swInfo.analog_in )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.analog_in )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit ) // Analoge Eingaenge
                    {
                    case 0x01: temp.append( "aI1" ); break;
                    case 0x02: temp.append( "aI2" ); break;
                    default:   temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein");
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // DigitalIn pruefen
        info.append( "Eingaenge Digital: " );
        if ( swInfo.digital_in )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.digital_in )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit ) // Digitale Eingaenge
                    {
                    case 0x01: temp.append( "dI1" ); break;
                    case 0x02: temp.append( "dI2" ); break;
                    case 0x04: temp.append( "dI3" ); break;
                    case 0x08: temp.append( "dI4" ); break;
                    case 0x10: temp.append( "dI5" ); break;
                    case 0x20: temp.append( "dI6" ); break;
                    case 0x40: temp.append( "dI7" ); break;
                    case 0x80: temp.append( "dI8" ); break;
                    default:   temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // Batterie
        info.append( boost::str( boost::format("Batterie: %1%\r\n") % (swInfo.battery ? "Ja" : "Nein") ) );

        // Ladeschaltung
        info.append( boost::str( boost::format("Ladeschaltung: %1%\r\n") % (swInfo.charge ? "Ja" : "Nein") ) );

        // Fingerprint pruefen
        info.append( "Fingerprint: " );
        if ( swInfo.fingerprint )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.fingerprint )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case BIOKEY_3000: temp.append( "Ja" ); break;
                    case BIOKEY_4000: temp.append( "Ja" ); break;
                    case BIOKEY_4020: temp.append( "Ja" ); break;
                    default:        temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // DCF77 pruefen
        info.append( "Funkuhr/Barcode: " );
        if ( swInfo.dcf77 )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.dcf77 )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case DCF77_HKW:           temp.append( "Funkuhr" ); break;
                    case DCF77_BARCODE:       temp.append( "Barcode Typ 1" ); break;
                    case DCF77_BARCODE_ZEBEX: temp.append( "Barcode Typ 2" ); break;
                    default:                  temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // GPS pruefen
        info.append( "GPS: " );
        if ( swInfo.gps )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.gps )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case GPS_FURUNO:       temp.append( "Ja, Typ 1" ); break;
                    case GPS_LEA:          temp.append( "Ja, Typ 2" ); break;
                    case GPS_MOUSE_LEA:    temp.append( "Ja, Typ 3" ); break;
                    default:               temp.append( "Undefiniert" ); break;
                    }

                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // MMC pruefen
        info.append( "MM Card: " );
        if ( swInfo.mmc )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.mmc )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case MMC_MEMORY: temp.append( "Speichererweiterung" ); break;
                    case MMC_DISK:   temp.append( "Wechseldatentraeger" ); break;
                    case MMC_USBHOST:temp.append( "USB-Memory-Stick" ); break;
                    default:         temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // Tastatur pruefen
        info.append( "Tastatur: " );
        if ( swInfo.keypad )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.keypad )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case KEYPAD_PZE:               temp.append( "9 Tasten" ); break;
                    case KEYPAD_BDE:               temp.append( "58 Tasten" ); break;
                    case KEYPAD_AE:                temp.append( "29 Tasten" ); break;
                    case KEYPAD_PZE_BASIC:         temp.append( "7 Tasten" ); break;
                    case KEYPAD_TIMEBOY:           temp.append( "25 Tasten" ); break;
                    case KEYPAD_FLEX:              temp.append( "20 Tasten" ); break;
                    case KEYPAD_TIMEBOY_MOBIL_PZE: temp.append( "3 Tasten" ); break;
                    default:                       temp.append( "Undefiniert" ); break;
                    }
                }

                if ( bit == 0x80 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();

        // Transponder pruefen
        info.append( "Transponderleser: " );
        if ( swInfo.transponderExtended )
        {
            for ( bit = 1; ; bit <<= 1 )
            {
                if ( bit & swInfo.transponderExtended )
                {
                    if ( temp.size() )
                        temp.append( ", " );

                    switch ( bit )
                    {
                    case TP_TSR30:       temp.append( "[EM4102, Hitag2]"); break;
                    case TP_TSR32:       temp.append( "[EM4102, EM4450, Hitag]"); break;
                    case TP_PRIMO100:    temp.append( "[Legic]"); break;
                    case TP_MIFARE:      temp.append( "[Mifare]"); break;
                    case TP_IBUTTON_EXT: temp.append( "[IButton (Extern)]"); break;
                    case TP_SMARTRELAIS: temp.append( "[SimonsVoss SmartRelais]"); break;
                    case TP_NEDAP:       temp.append( "[Nedap]"); break;
                    case TP_IBUTTON_INT: temp.append( "[IButton (Intern)]"); break;
                    case TP_HID_PROXKEY: temp.append( "[HID ProxKey]"); break;
                    case TP_MULTI_ISO_OEM_READER: temp.append( "[Mifare ISO14443/ISO15693]"); break;
                    case TP_LEGIC_ADVANT: temp.append( "[Legic Advant]"); break;
                    case TP_TSR21:       temp.append( "[EM4102, EM4450, Hitag]"); break;
                    case TP_HID_ICLASS:  temp.append( "[iClass, Wiegand, Clock Data]"); break;
                    case TP_TB_MIFARE:   temp.append( "[Timeboy Mifare]"); break;
                    case TP_TB_LEGIC:    temp.append( "[Timeboy Legic]"); break;
                    case TP_SMARTRELAIS_G2: temp.append( "[SimonsVoss SmartRelais G2]"); break;
                    case TP_MIFARE_NFC_OEM_READER: temp.append( "[Mifare ISO14443]"); break;
                    case TP_IDTRONIC_UHF: temp.append( "[ID Tronic UHF]"); break;
                    default:             temp.append( "[Undefiniert]"); break;
                    }
                }

                if ( bit == 0x80000000 )
                    break;
            }
        }
        else
        {
            temp.append( "Nein" );
        }
        temp.append( "\r\n" );
        info.append( temp );
        temp.clear();
    }

    return res;
}
