/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/connection_base.h"

#if !defined(_MSC_VER) && !defined(__MINGW32__)
#include <sys/ioctl.h>
#endif

#if defined(_MSC_VER) && !defined(snprintf)
#define snprintf _snprintf
#endif

#include <string>
#include <cstring>
#include <cstdio>
#include <cstdarg>
#include <ctime>
#include <memory>
#include <thread>
#include <sstream>

#include <boost/format.hpp>

Datafox::Communication::Trace_base::Trace_base( int id )
{
    m_id = id;
    // Tracemeldungen
    m_mode  = Trace_mode::off;
    // Tracefilename
    m_file_path = boost::str( boost::format( "DFComDLL_%1%.log" ) % m_id );

#ifdef _DEBUG
    // Nur im Debugmodus standardmaessig aktiv.
    set_mode();
#endif
}

// ##################################################################
/**
* @brief Hauptlogfunktion der Verbindungsklassen.
*/
void Datafox::Communication::Trace_base::print( const char *format, va_list args )
{
    if ( get_mode() >= Trace_mode::on )
    {
        if (/*std::strstr(format, "INFO(1)") != NULL ||
            std::strstr(format, "INFO(2)") != NULL ||*/
            std::strstr(format, "INFO(3)") != NULL || /*
            std::strstr(format, "WARNING(1)") != NULL ||
            std::strstr(format, "WARNING(2)") != NULL ||*/
            std::strstr(format, "WARNING(3)") != NULL /*||
            std::strstr(format, "ERROR(1)") != NULL ||
            std::strstr(format, "ERROR(2)") != NULL ||
            std::strstr(format, "ERROR(3)") != NULL*/ )
        {
            return;
        }

        va_list args2;
        va_copy( args2, args );
        std::string buffer( vsnprintf( nullptr, 0, format, args ) + 1, '\0' );
        vsnprintf( &buffer[0], buffer.size(), format, args2 );
        va_end( args2 );

        // Eine ggf. mitberechnete Nullterminierung entfernen.
        if ( buffer[buffer.size() - 1] == '\0' )
        {
            buffer.resize( buffer.size() - 1 );
        }

        std::size_t pos;
        if ( (pos = buffer.find( ":INFO(" )) != std::string::npos || 
            (pos = buffer.find( ":WARNING(" )) != std::string::npos || 
            (pos = buffer.find( ":ERROR(" )) != std::string::npos )
        {
            // Uhrzeit des Logeintrages
            time_t t = time( NULL );
            tm *ltm = localtime( &t );
            unsigned short msec = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
            SYSTEMTIME st;
            GetLocalTime( &st );
            msec = st.wMilliseconds;
#endif
            // Uhrzeit einfuegen
            std::string temp( 100, '\0' );
            snprintf( &temp[0], temp.size(), ":%02i:%02i:%02i.%03i", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, msec );
            buffer.insert( pos, temp.c_str() );

            // ThreadID einfuegen
            std::stringstream ss_ids;
            ss_ids << ":(" << get_id() << ", " << std::this_thread::get_id() << ")";
            buffer.insert( pos, ss_ids.str().c_str() );
        }

#if !defined(_MSC_VER) && !defined(__MINGW32__)
        if ( buffer[strlen(buffer.c_str()) - 1] == '\n' )
            buffer.append( "\r\n" );
#endif

        FILE* fp = fopen( m_file_path.data(), "a" );
        if ( fp == nullptr )
            return;

        fwrite( buffer.data(), 1, buffer.size(), fp );
        fclose( fp );

#ifdef _DEBUG
#if defined(_MSC_VER) || defined(__MINGW32__)
        OutputDebugStringA( buffer.data() );
#endif
#endif
    }
}

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connection_id Nummer der Verbindungsklasse.
*/
Datafox::Communication::Connection_base::Connection_base( int id ) 
     : m_trace( std::make_shared<Trace_base>( id ) )
{
    // Verbindungsnummer zuweisen
    m_id = id;

    // Membervars initialisieren
    m_type    = Connection_type::undefined;
    m_serial_port  = INVALID_HANDLE_VALUE;
    m_socket  = INVALID_SOCKET;
    m_value   = 0;
    m_read_timeout = 0;
#if !defined(_MSC_VER) && !defined(__MINGW32__)
    std::fill_n( (uint8_t *)&m_flock, sizeof( struct flock ), 0 );
    std::fill_n( (uint8_t *)&m_termio, sizeof( struct termio ), 0 );
#endif

    // Statistik
    m_size_written = 0;
    m_size_read = 0;
}

// ##################################################################
/**
* @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
*
*/
Datafox::Communication::Connection_base::~Connection_base() 
{
    // Schliessen der Verbindung, wenn nicht bereits geschehen.
    close();
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

// ##################################################################
/**
* @brief Verwirft anstehende Zeichen der Schnittstelle.
*/
void Datafox::Communication::Connection_base::flush()
{
    trace( "%s(%d):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

    unsigned char buffer[0x100];
    std::size_t read_count = 0xFF;

    // eventuell anstehende Datenpakete verwerfen
    for ( int i = 0; (i < 5) && read_count; ++i )
    {
        if ( read_data( buffer, sizeof( buffer ) - 1, read_count ) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%d):INFO(2): %s() [ Abbruch nach %d durchläufen. ]\n", __FILE__, __LINE__, __func__, i );
#ifdef _DEBUG
            for ( int z = 0; z < read_count; ++z ) 
            {
                trace( "%02X ", buffer[z] );
            }

            trace( "\n" );
#endif
            break;
        }

        trace( "%s(%d):INFO(2): %s() [ Verworfene Anzahl Zeichen (%u Byte) ]\n", __FILE__, __LINE__, __func__, read_count );
#ifdef _DEBUG
        for ( int z = 0; z < read_count; ++z ) 
        {
            trace( "%02X ", buffer[z] );
        }

        trace( "\n" );
#endif
    }
}

// ##################################################################
/**
* @brief Oeffnen der Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::open( Connection_type connection_type, const std::string &connection_string, uint32_t connection_value, uint32_t communication_read_timeout )
{
    trace( "%s(%d):INFO(3): %s( %d, %s, %u, %u )\n", __FILE__, __LINE__, __func__, connection_type, connection_string.c_str(), connection_value, communication_read_timeout );

    int res = IV_ERROR_FAIL;

    switch ( connection_type )
    {
    case Connection_type::serial:

        if ( communication_read_timeout == 0 )
        {
            // Von einem Standardtimeout 800ms ausgehen.
            if ( m_read_timeout == 0 )
            {
                communication_read_timeout = 800;
            }
            else
            {
                communication_read_timeout = m_read_timeout; // Zuvor gewaehlte.
            }

        }
        else if ( communication_read_timeout < 500 )
        { 
            // Mindesttimeout sicherstellen.
            trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 500 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
            communication_read_timeout = 500;
        }
        else if ( communication_read_timeout > 60000 )
        {
            // Maximal 1 Minute als Timeout zulassen.
            trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
            communication_read_timeout = 60000;
        }

        if ( connection_value == 0 )
        {
            // Von einer Standardbaudrate 38400bd ausgehen.
            if ( m_value == 0 )
            {
                connection_value = 38400;
            }
            else 
            {
                connection_value = m_value; // Zuvor gewaehlte.
            }
        }

        if ( open_serial( connection_string, connection_value, communication_read_timeout ) == IV_ERROR_SUCCESS )
        {
            m_type = Connection_type::serial;
            m_value = connection_value;
            m_read_timeout = communication_read_timeout;
            m_string = connection_string;

            res = IV_ERROR_SUCCESS;
        }
        break;

    case Connection_type::socket:
        // Mindesttimeout sicherstellen.
        if ( communication_read_timeout < 2000 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
            communication_read_timeout = 2000;
        }
        else if ( communication_read_timeout > 60000 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
            communication_read_timeout = 60000;
        }

        if ( open_socket( connection_string, (uint16_t)connection_value, communication_read_timeout ) == IV_ERROR_SUCCESS )
        {
            m_type = Connection_type::socket;
            m_value = connection_value;
            m_read_timeout = communication_read_timeout;
            m_string = connection_string;

            res = IV_ERROR_SUCCESS;
        }
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, connection_type );
        return res;
    }

    return res;
}

// ##################################################################
/**
* @brief Ermitteln ob die Kommunikationsschnittstelle geoeffnet ist.
*/
bool Datafox::Communication::Connection_base::is_open()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Bei Socketverbindung sicherstellen, dass diese aktuell noch besteht.
    if ( get_type() == Connection_type::socket )
    {
#if _MSC_VER 
#if !defined(MYCONST)
#define MYCONST const
#endif
#else
#define MYCONST
#endif
        struct timeval seltime;
        fd_set readfds;
        char buffer = 0;

        int res;

        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = 0;
        seltime.tv_usec = 500;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
            close();
        }
        else if ( res ) // Liegen Daten an?
        {
            if ( (res = recv( m_socket, (char*)&buffer, 1, MSG_PEEK )) == SOCKET_ERROR )
            {
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                close();
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                close();
            }
        }
        /*
        int res;
        char buffer = '?';
        // Daten senden
        if ( (res = send( _socket, "?", 1, MSG_OOB )) == SOCKET_ERROR )
        {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
        close();
        }

        // Wurden alle Daten versendet?
        if ( res == 0 )
        {
        trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. id: %d ]\n", __FILE__, __LINE__, __func__, getId() );
        close();
        }
        */
    }

    return (get_type() != Connection_type::undefined);
}

// ##################################################################
/**
* @brief Schliessen der Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::close()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    switch ( get_type() )
    {
    case Connection_type::undefined:
        return IV_ERROR_SUCCESS;

    case Connection_type::socket:
        close_socket();
        break;

    case Connection_type::serial:
        close_serial();
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, get_type() );
        return IV_ERROR_FAIL;
    }

    m_type = Connection_type::undefined;
    m_string = "";

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Zuweisen einer vorhandenen Kommunikationsschnittstelle.
* 
*/
int Datafox::Communication::Connection_base::attach( Connection_type connection_type, HANDLE connection_handle, uint32_t communication_read_timeout )
{
    trace("%s(%d):INFO(3): %s( %d, %p, %u )\n", __FILE__, __LINE__, __func__, connection_type, connection_handle, communication_read_timeout);

    bool was_undefined = (get_type() == Connection_type::undefined);

    if ( get_type() == Connection_type::undefined
        || (get_type() == connection_type && connection_type == Connection_type::serial && connection_handle == m_serial_port)
        || (get_type() == connection_type && connection_type == Connection_type::socket && connection_handle == (HANDLE)m_socket) )
    {
        switch ( connection_type )
        {
        case Connection_type::serial:
            m_type = Connection_type::serial;
            m_serial_port = (HANDLE)connection_handle;
            m_read_timeout   = communication_read_timeout;
            break;

        case Connection_type::socket:
            if ( communication_read_timeout < 2000 )
            {
                trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
                communication_read_timeout = 2000;
            }
            else if ( communication_read_timeout > 60000 )
            {
                trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communication_read_timeout );
                communication_read_timeout = 60000;
            }

            m_type   = Connection_type::socket;
            m_socket = (SOCKET)connection_handle;
            m_read_timeout     = communication_read_timeout;
            break;

        default:
            trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, connection_type );
            return IV_ERROR_FAIL;
        }

        if ( was_undefined ) // Kanal Zugewiesen
            m_string = "ATTACH";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%d):ERROR(1): %s() [ Zuweisen des Kanals nicht moeglich, da nicht Undefiniert. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;

}

// ##################################################################
/**
* @brief Freigeben der aktuellen Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::detach()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    if ( get_type() == Connection_type::serial || m_string == "ATTACH" )
    {
        m_type = Connection_type::undefined;
        m_value = 0;
        m_read_timeout = 0;
        m_serial_port = INVALID_HANDLE_VALUE;
        m_socket = INVALID_SOCKET;

        // Kommunikationsstring loeschen
        m_string = "";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%d):ERROR(1): %s() [ Freigabe des Handle nicht moeglich, da kein serial-port. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;
}

// ##################################################################
/**
* @brief Oeffnen einer seriellen Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::open_serial( const std::string &serial_port_name, uint32_t baud_rate, uint32_t communication_read_timeout )
{
    trace( "%s(%d):INFO(3): %s( %s, %u, %u )\n", __FILE__, __LINE__, __func__, serial_port_name.c_str(), baud_rate, communication_read_timeout );

    int rc;

    if ( get_type() != Connection_type::undefined )
    {
        // Verbindungs-Id bereits geoeffnet
        trace( "%s(%d):ERROR(1): %s() [ Verbindungs-Id bereits geoeffnet. ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

#if defined(_MSC_VER) || defined(__MINGW32__)

    if ( (m_serial_port = CreateFileA(
        serial_port_name.data(),        // pointer to name of the file 
        (GENERIC_READ | GENERIC_WRITE), // access (read-write) mode 
        0,                              // share mode 
        NULL,                           // pointer to security descriptor 
        OPEN_EXISTING,                  // how to create 
        0,                              // file attributes 
        0                               // handle to file with attributes to copy  
        )) == INVALID_HANDLE_VALUE )
    { 
        trace( "%s(%d):ERROR(1): %s() [ CreateFile fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    errno = 0;

    if ( (rc = ::open( serial_port_name.data(), O_RDWR | O_NDELAY )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ open fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    m_serial_port = rc;

    if ( (rc = fcntl( m_serial_port, F_SETFL, O_RDWR )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ fcntl fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    m_flock.l_type    = F_WRLCK;  // lock stream for read and write
    m_flock.l_whence  = 0;        // lock whole file -> 0,0,0
    m_flock.l_start   = 0;
    m_flock.l_len     = 0;

    if ( (rc = fcntl( m_serial_port, F_SETLK, &m_flock )) < 0 )
    {
        trace( "%s(%d):WARINING(1): %s() [ fcntl (file locking) fail ]\n", __FILE__, __LINE__, __func__ );
        //return IV_ERROR_FAIL;
    }
#endif

    if ( (rc = change_state_serial( m_serial_port, baud_rate, communication_read_timeout )) != IV_ERROR_SUCCESS )
    {
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Aendern der aktuellen Parameter zur seriellen Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::change_state_serial( HANDLE connection_serial_port, uint32_t baud_rate, uint32_t communication_read_timeout )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u )\n", __FILE__, __LINE__, __func__, connection_serial_port, baud_rate, communication_read_timeout);

#if defined(_MSC_VER) || defined(__MINGW32__)
    DCB          dcb;
    COMMTIMEOUTS ctout;
    DWORD        cerr = 0;

    memset( &dcb, 0, sizeof( dcb ) );
    dcb.DCBlength = sizeof( dcb );
    dcb.fBinary   = 1;

    dcb.BaudRate  = (DWORD)baud_rate;
    dcb.ByteSize  = 8;
    dcb.Parity    = NOPARITY;
    dcb.StopBits  = ONESTOPBIT;

    if ( !SetCommState( connection_serial_port, &dcb ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ SetCommState fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        close_serial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connection_serial_port == m_serial_port )
    {
        // Akutuelle Baudrate anpassen.
        m_value = baud_rate;
    }

    ctout.ReadIntervalTimeout         = 0;
    ctout.ReadTotalTimeoutMultiplier  = 0;
    ctout.ReadTotalTimeoutConstant    = (DWORD)communication_read_timeout;
    ctout.WriteTotalTimeoutConstant   = 0;
    ctout.WriteTotalTimeoutMultiplier = 0;

    if ( !SetCommTimeouts( connection_serial_port, &ctout ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ SetCommTimeouts fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connection_serial_port == m_serial_port )
    {
        // Aktuellen Timeout anpassen
        m_read_timeout = communication_read_timeout;
    }

    // unsolved: fuer modems + RS485 notwendig !?!
    if ( !EscapeCommFunction( connection_serial_port, SETRTS ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }
    if ( !EscapeCommFunction( connection_serial_port, SETDTR ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    if ( !ClearCommError( connection_serial_port, &cerr, NULL ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ ClearCommError fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

#else
    int rc;
    int baud_mode, char_mode, pari_check, stop_mode;
    int termio_cm;

    switch( baud_rate )
    {
    case 9600:
        baud_mode = B9600;
        break;

    case 19200:
        baud_mode = B19200;
        break;

    case 38400:
        baud_mode = B38400;
        break;
    case 57600:
        baud_mode = B57600;
        break;
    case 115200:
        baud_mode = B115200;
        break;
    default:
        return IV_ERROR_FAIL;
    }

    char_mode = CS8;
    pari_check = 0;
    stop_mode = 0;

    m_termio.c_iflag = 0;
    m_termio.c_oflag = 0;
    m_termio.c_cflag = 0;
    m_termio.c_lflag = 0;
    m_termio.c_line  = 0;
    memset( (void *)m_termio.c_cc, 0, NCCS );

    m_termio.c_cflag |= HUPCL;        // hangup line when close/exit(DTR of)
    m_termio.c_cflag |= CLOCAL;       // local/direct connection

    m_termio.c_cflag |= baud_mode;     // set baudrate
    m_termio.c_cflag |= char_mode;     // set databits
    m_termio.c_cflag |= pari_check;    // set parity
    m_termio.c_cflag |= stop_mode;     // set stopbits

    m_termio.c_cflag |= CREAD;        // enable input queue

    m_termio.c_cc[VMIN]  = 1;         // read waiting for 1 char

    if ( (rc = ioctl( connection_serial_port, TCSETA, &m_termio )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( connection_serial_port, TCFLSH, TCIOFLUSH )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( connection_serial_port, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    if ( !(termio_cm & TIOCM_RTS) )
        termio_cm |= TIOCM_RTS;

    if ( !(termio_cm & TIOCM_DTR) )
        termio_cm |= TIOCM_DTR;

    if ( (rc = ioctl( connection_serial_port, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        close_serial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connection_serial_port == m_serial_port )
    {
        // Akutuelle Baudrate und Timeout anpassen.
        m_value = baud_rate;
        m_read_timeout = communication_read_timeout;
    }
#endif

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schliessen der seriellen Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::close_serial()
{
    int res = IV_ERROR_SUCCESS;
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER) || defined(__MINGW32__)
    try
    {
        if ( !CancelIo( m_serial_port ) )
        {
            trace( "%s(%d):ERROR(1): %s() [ CancelIo fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }

    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( m_serial_port, CLRRTS ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }
    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( m_serial_port, CLRDTR ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }

    try
    {
        if ( !CloseHandle( m_serial_port ) )
        {
            trace( "%s(%d):ERROR(1): %s() [ CloseHandle fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }
#else
    int rc;
    int termio_cm;

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( m_serial_port, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    if ( (termio_cm & TIOCM_RTS) )
        termio_cm &= ~TIOCM_RTS;

    if ( (termio_cm & TIOCM_DTR) )
        termio_cm &= ~TIOCM_DTR;

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( m_serial_port, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    ::close( m_serial_port );
#endif
    m_serial_port = INVALID_HANDLE_VALUE;

    return res;
}

// ##################################################################
/**
* @brief Lesen aus der seriellen Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::read_serial( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials )
{
    trace("%s(%d):INFO(3): %s( %p, %u, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, read_count, read_in_count, read_trials);

    std::size_t res;
    uint8_t trials = 0;
    read_in_count = 0;
#if !defined(_MSC_VER) && !defined(__MINGW32__)
    struct timeval seltime;
    fd_set readfds;
    errno = 0;
#endif

    // Daten emfpangen
    for ( ; (read_count != read_in_count) && (trials < read_trials); ++trials )
    {
        res = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( !ReadFile( m_serial_port, buffer + read_in_count, DWORD(read_count - read_in_count), (LPDWORD)&res, nullptr ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%d):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %u ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }

            trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, read_count, read_in_count );
            return -IV_ERROR_RECEIVE;
        }
#else
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_serial_port, (fd_set FAR *) &readfds );
        seltime.tv_sec  = m_read_timeout / 1000;
        seltime.tv_usec = (m_read_timeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (struct timeval *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, read_count, read_in_count );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        { 
            if ( (res = read( m_serial_port, buffer + read_in_count, read_count - read_in_count )) < 0 )
            {
                trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, read_count, read_in_count );
                return -IV_ERROR_RECEIVE;
            }
        }
#endif
        if ( res == 0 )
        {
#if defined(_MSC_VER) || defined(__MINGW32__)
            trace( "%s(%d):WARNING(1): %s() [ Leseversuch ohne Daten. (WinError: %u)]\n", __FILE__, __LINE__, __func__, GetLastError() );
#else
            trace( "%s(%d):WARNING(1): %s() [ Leseversuch ohne Daten. ]\n", __FILE__, __LINE__, __func__ );
#endif
            break;
        }
        else if ( res < (int)(read_count - read_in_count) )
        {
            trace( "%s(%d):WARNING(1): %s() [ Nachlesen erforderlich. trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, trials, res, read_count - read_in_count, read_in_count );
        }

        read_in_count += res;
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        m_size_read += res;
    }

    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( read_count != read_in_count )
    {
        trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, read_count, read_in_count );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schreiben in die serielle Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::write_serial( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, write_count, write_out_count, write_trials );

#if !defined(_MSC_VER) && !defined(__MINGW32__)
    errno = 0;
#endif

    std::size_t res = IV_ERROR_SUCCESS;
    uint8_t trials = 0;
    write_out_count = 0;

    for ( ; (write_count != write_out_count) && (trials < write_trials); ++trials )
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( !WriteFile( m_serial_port, buffer + write_out_count, DWORD(write_count - write_out_count), (LPDWORD)&res, nullptr ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%d):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %u ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }
#else

        if ( (res = write( m_serial_port, (unsigned char*)buffer + write_out_count, write_count - write_out_count )) < 0 )
        {
#endif
#if defined(_MSC_VER) || defined(__MINGW32__)
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, trials, write_count, write_out_count, GetLastError() );
            if ( GetLastError() == ERROR_ALREADY_EXISTS )
            {            
                // 2014-04-03
                // Workaround um bei USB-COM-Ports ein blockieren zu verhindern.
                std::this_thread::sleep_for( std::chrono::seconds( 3 ) );
            }
#else
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, trials, write_count, write_out_count );
#endif
            return -IV_ERROR_TRANSMIT;
        }

        // Wurden alle Zeichen versendet?
        if ( res == 0 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. ]\n", __FILE__, __LINE__, __func__ );
            break;

        }
        else if ( res < (write_count - write_out_count) )
        {		
            trace( "%s(%d):WARNING(1): %s() [ Nachsenden erforderlich. trial: %u, waren %u, gefordert %u, bereits %u]\n", __FILE__, __LINE__, __func__, trials, res, write_count - write_out_count, write_out_count );
        }

        write_out_count += res;
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        m_size_written += res;
    }

    // Wurden nicht alle Zeichen versendet?
    if ( write_count != write_out_count )
    {
        trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, trials, write_count, write_out_count );
        return -IV_ERROR_TRANSMIT;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Oeffnen einer TCP/IP Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::open_socket( const std::string &host_name, uint16_t port_number, uint32_t communication_read_timeout )
{
    trace( "%s(%d):INFO(3): %s( %s, %u, %u )\n", __FILE__, __LINE__, __func__, host_name.c_str(), port_number, communication_read_timeout );

    SOCKADDR_IN sin;
    memset( &sin, 0, sizeof(sin) );
    sin.sin_family = AF_INET;
    sin.sin_port = htons( port_number );
    sin.sin_addr.s_addr = inet_addr( host_name.data() );

    // Kanal bereits belegt?
    if ( get_type() != Connection_type::undefined )
        return IV_ERROR_FAIL;

    if ( sin.sin_addr.s_addr == INADDR_NONE )
    { 
        // Adresse ueber DNS aufloesen
        HOSTENT *he;
        he = gethostbyname( host_name.data() );
        if ( !he )
        {
            trace( "%s(%d):ERROR(1): %s() [ gethostbyname fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
            return IV_ERROR_FAIL;
        }

        sin.sin_addr.s_addr = *((unsigned int*)he->h_addr);
    }

    if ( (m_socket = ::socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET )
    {
        trace( "%s(%d):ERROR(1): %s() [ socket fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }

    if ( connect( m_socket, (SOCKADDR*)&sin, sizeof(sin) ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ connect fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        close_socket();
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schliessen der TCP/IP Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::close_socket()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( closesocket( m_socket ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ closesocket fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    if ( shutdown( m_socket, 2 ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ shutdown fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
    }
    if ( ::close( m_socket ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ close fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#endif

    m_socket = INVALID_SOCKET;

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen aus der TCP/IP Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::read_socket( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, read_count, read_in_count, read_trials );

#if defined(_MSC_VER)
#if !defined(MYCONST)
#define MYCONST const
#endif
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    std::size_t res;
    uint8_t trials = 0;
    uint8_t lTrial = (uint8_t)read_count;

    read_in_count = 0;

    // Daten emfpangen
    for ( ; (read_count != read_in_count) && (trials < read_trials); ++trials )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = m_read_timeout / 1000;
        seltime.tv_usec = (m_read_timeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, read_count, read_in_count, WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {
            if ( (res = recv( m_socket, (char *)buffer + read_in_count, int(read_count - read_in_count), 0 )) == SOCKET_ERROR )
            {
#if defined(_MSC_VER) || defined(__MINGW32__)
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%d):ERROR(1): %s() [ id: %d, recv error: %u ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, read_count, read_in_count, WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, read_count, read_in_count );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
            else if ( res < (read_count - read_in_count) )
            {
                trace( "%s(%d):WARNING(1): %s() [ Nachlesen erforderlich. id: %d, trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, res, read_count - read_in_count, read_in_count );
                // Fuer Kommunikation ueber Com-Server mit Mehrfachdocking viele Teilpackete
                if ( lTrial )
                    trials--;
                else
                    lTrial = uint8_t(read_count - read_in_count);

                lTrial--;
            }
        }
        else
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gelesen. id: %d ]\n", __FILE__, __LINE__, __func__, get_id() );
            break;
        }

        read_in_count += res;
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        m_size_read += res;
    }

    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( read_count != read_in_count )
    {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, read_count, read_in_count );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schreiben in die TCP/IP Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::write_socket( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, write_count, write_out_count, write_trials );

    std::size_t res;
    uint8_t trials = 0;

    write_out_count = 0;

    for ( ; (write_count != write_out_count) && (trials < write_trials); ++trials )
    {
        // Daten senden
        if ( (res = send( m_socket, (char*)buffer + write_out_count, int(write_count - write_out_count), 0 )) == SOCKET_ERROR )
        {
#if defined(_MSC_VER) || defined(__MINGW32__)
            if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
            {
                trace( "%s(%d):ERROR(1): %s) [ id: %d, send error: %u ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
#endif
            trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, writeCount %u, writeOutCount: %u, Error: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, write_count, write_out_count, WSAGetLastError() );
            return -IV_ERROR_TRANSMIT;
        }

        // Wurden alle Daten versendet?
        if ( res == 0 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. id: %d ]\n", __FILE__, __LINE__, __func__, get_id() );
            break;
        }
        else if ( res < (write_count - write_out_count) )
        {
            trace( "%s(%d):WARNING(1): %s() [ Nachsenden erforderlich. id: %d, trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, res, write_count - write_out_count, write_out_count );
        }

        write_out_count += res;
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        m_size_written += res;
    }

    // Wurden nicht alle Daten im vorgegebenen Rahmen versendet?
    if ( write_count != write_out_count )
    {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, writeCount %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, get_id(), trials, write_count, write_out_count );
        return -IV_ERROR_TRANSMIT;
    }

    // Ausfuehrung ist erfolgt.
    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Verwerfen der Daten an der TCP/IP Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::flush_socket()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER)
#define MYCONST const
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    int res;
    char buffer[1024];

    // Daten emfpangen
    for ( ;; )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = 0;
        seltime.tv_usec = 500000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {
            if ( (res = recv( m_socket, buffer, sizeof(buffer), 0 )) == SOCKET_ERROR )
            {
#if defined(_MSC_VER) || defined(__MINGW32__)
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%d):ERROR(1): %s() [ id: %d, recv error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, get_id(), WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d ]\n", __FILE__, __LINE__, __func__, get_id() );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
        }
        else
        {
            break;
        }
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen aus der Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::read_data( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, read_count, read_in_count, read_trials );

    int result = IV_ERROR_SUCCESS;
    switch ( get_type() )
    {
    case Connection_type::serial:
        result = read_serial( buffer, read_count, read_in_count, read_trials );
        break;

    case Connection_type::socket:
        result = read_socket( buffer, read_count, read_in_count, read_trials );
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ unknown connection type: %d ]\n", __FILE__, __LINE__, __func__, get_type() );
        result = -IV_ERROR_RECEIVE;
    }

    return result;
}

// ##################################################################
/**
* @brief Schreiben in die Kommunikationsschnittstelle.
*/
int Datafox::Communication::Connection_base::write_data( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, std::uint8_t write_trials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, write_count, write_out_count, write_trials );

    int result = IV_ERROR_SUCCESS;
    switch ( get_type() )
    {
    case Connection_type::serial:
        result = write_serial( buffer, write_count, write_out_count, write_trials );
        break;

    case Connection_type::socket:
        result = write_socket( buffer, write_count, write_out_count, write_trials );
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ unknown connection type: %d ]\n", __FILE__, __LINE__, __func__, get_type() );
        result = -IV_ERROR_TRANSMIT;
    }

    return result;
}

