/**
* @file  file_device_zip.cpp
* @brief Implementierung der Klasse FileDeviceZip.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_device_zip.h"
#include <sys/types.h>
#include <sys/stat.h>

#include <list>
#include <cstdio>
#include <memory>

#if defined(_MSC_VER)
#include <filesystem>
namespace fs = std::experimental::filesystem;
#endif

FileDeviceZip::FileDeviceZip()
{
    // Initialisieren
    resetData();

    // Auch den Error mitresetten.
    resetError();
}

FileDeviceZip::~FileDeviceZip()
{
}

void FileDeviceZip::resetData()
{
    // Basis Reset
    FileDevice::resetData();
}

void FileDeviceZip::resetError()
{
    // Basis Reset
    FileDevice::resetError();
    // Fehlertext loeschen
    m_errorExtended.clear();
}

void FileDeviceZip::setRecentZipError(const std::string &filePath, int lineNumber)
{
    std::string text;
#if defined(_MSC_VER) || defined(__MINGW32__)
    // Fehlertext ermitteln.
    text.resize(255);
    FormatZipMessage( ZR_RECENT, &text[0], static_cast<unsigned int>( text.size() ) );
#endif
    setErrorText( text );
    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % filePath % lineNumber % text ) );
}

void FileDeviceZip::setErrorTextExtended( const std::string &text )
{
    m_errorExtended = text;
}

void FileDeviceZip::appendErrorTextExtended( const std::string &text )
{
    m_errorExtended.append( text );
}

std::string FileDeviceZip::errorTextExtended()
{
    return m_errorExtended;
}

int FileDeviceZip::CreateZipFromDirectory(char *path, char *fileZIP, char *defaultDFL)
{
    int loaded = 0;

#if defined(_MSC_VER) || defined(__MINGW32__)
    char szZipFile[260], szFile[260], szIniFile[260];
    char szSearchFile[260];
    intptr_t fp;
    struct _finddata_t finddata;
    std::string version;
    std::list<std::string> versions;
    std::string timestamp;
    std::string comment;
    int fileCount = 0;
    char *md5 = NULL;

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Speicher initialisieren
    memset(szZipFile, 0, sizeof(szZipFile));
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    memset(szSearchFile, 0, sizeof(szSearchFile));

    // Je nach Angabe Basispfad mit berücksichtigen
    if (fileZIP[1] == ':') {

        strcpy(szZipFile, fileZIP);
    } else {

        sprintf(szZipFile, "%s\\%s", path, fileZIP);
    }

    sprintf(szIniFile, "%s\\%s", path, "SWI.ini");
    ::remove(szIniFile); // Vorhandene Initialisierungsdatei loeschen

    sprintf(szSearchFile, "%s\\*.hex", path);
    // Passende Dateien fuer Muster?
    if ((fp = _findfirst(szSearchFile, &finddata)) != -1) {

        // Vorhandenes Archiv loeschen
        ::remove(szZipFile); 
        // Zipdatei erzeugen
        HZIP hZip = NULL;
        if ((hZip = CreateZip(szZipFile, 0)) == NULL) {

            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }

        // Builddatum, wenn nicht vorhanden, eintragen
        if (timestamp.size() != 19) {

            char tmp[20];
            time_t ltime;
            struct tm *ltm;

            time(&ltime);
            ltm = localtime(&ltime);
            strftime(tmp, sizeof(tmp), "%Y-%m-%d %H:%M:%S", ltm);
            timestamp = std::string( tmp );
        }

        // Alle Dateien durchlaufen.
        do {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s\\%s", path, finddata.name);

            // Umgebung zuruecksetzen
            resetData();
            // Fehlerobjekt zuruecksetzen
            resetError();

            // Dateidaten laden
            if (!load(szFile)) {

                // Eigenen Fehlertext setzen.
                setErrorText( boost::str( boost::format( "Fehler (%1%) beim Ladeversuch der Datei [ %2% ]" ) %  static_cast<int>( errorState() ) % szFile ) );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                goto error_label;
            }

            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            if (GetHeaderSoftwareInfo().size) {

                // MD5 der Datei in Initialisierungsdatei uebernehmen
                char *md5 = NULL;
                if ((md5 = MD5File(szFile)) == NULL) {

                    // Datei kann nicht uebernommen werden.
                    // Eigenen Fehlertext setzen.
                    setErrorText( "Erstellen des MD5(1) schlug fehl." );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                    goto error_label;
                }

                m_iniFile.WriteString(finddata.name, "MD5", md5);
                free(md5);

                // Version zur Datei in Initialisierungsdatei uebernehmen.
                GetHeaderVersionFirmware( version );
                if ( version.length() > 11 ) // Version nur als 00.00.00.00 Format zulassen, ggf. Endung ZK, PZE etc. wegschneiden.
                    version.resize( 11 );

                m_iniFile.WriteString(finddata.name, "Version", version.data());

                // Versionen fuer Archivini halten.
                // Pruefung ob diese Version bereits eingetragen ist.
                if ( std::find( versions.begin(), versions.end(), version ) == versions.end() )
                    versions.push_back( version );

                // Softwareinfo zur Datei in Initialisierungsdatei uebernehmen.
                std::string tmp;
                StructSoftwareInfo swInfo = GetHeaderSoftwareInfo();
                SoftwareInfoStructToString(tmp, swInfo);
                m_iniFile.WriteString(finddata.name, "SWI", tmp.data());

                // Datei zu Archiv hinzufuegen
                if (ZipAdd(hZip, finddata.name, szFile) != ZR_OK) {
                    // Fehlertext von Zip-API setzen.
                    setRecentZipError( __FILE__, __LINE__ );
                    goto error_label;
                }

                fileCount++;
#if !defined(__MINGW32__)
                // Dateien auch als IFF-Format eintragen.
                if ( SoftwareVersion( version ) > "04.03.00.00" )
                {
                    fs::path iff_file_path( szFile );
                    iff_file_path.replace_extension( "iff" );

                    if ( fs::exists( iff_file_path ) )
                    {
                        // MD5 der Datei in Initialisierungsdatei uebernehmen
                        char *md5 = nullptr;
                        if ( (md5 = MD5File( iff_file_path.string().c_str() )) == nullptr )
                        {
                            // Datei kann nicht uebernommen werden.
                            // Eigenen Fehlertext setzen.
                            setErrorText( "Erstellen des MD5(1) schlug fehl." );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                            goto error_label;
                        }

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "MD5", md5 );
                        free( md5 );

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "Version", version.data() );

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "SWI", tmp.data() );

                        // Datei zu Archiv hinzufuegen
                        if ( ZipAdd( hZip, iff_file_path.filename().string().c_str(), iff_file_path.string().c_str() ) != ZR_OK )
                        {
                            // Fehlertext von Zip-API setzen.
                            setRecentZipError( __FILE__, __LINE__ );
                            goto error_label;
                        }

                        fileCount++;
                    }
                }
#endif
            }

        } while (_findnext(fp, &finddata) != -1);

        // Textdaten fuer Firmware verfuegbar?
        if (defaultDFL != NULL) {

            // MD5 der Datei in Initialisierungsdatei uebernehmen
            char *md5 = NULL;
            if ((md5 = MD5String(defaultDFL)) == NULL) {

                // Datei kann nicht uebernommen werden.
                // Eigenen Fehlertext setzen.
                setErrorText( "Erstellen des MD5(1) schlug fehl." );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                goto error_label;
            }

            m_iniFile.WriteString("Default.dfl", "MD5", md5);
            free(md5);

            // Datei mit Textdaten schreiben
            if (ZipAdd(hZip, "Default.dfl", defaultDFL, (unsigned int)strlen(defaultDFL)) != ZR_OK) {
                // Fehlertext von Zip-API setzen.
                setRecentZipError( __FILE__, __LINE__ );
                goto error_label;
            }
        }

        // Initialisierungsdatei schreiben
        m_iniFile.WriteIniFile(szIniFile);
        m_iniFile.CloseIniFile();
        // Datei zu Archiv hinzufuegen
        if (ZipAdd(hZip, "SWI.ini", szIniFile) != ZR_OK) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }

        // MD5 der Initialisierungsdatei als Name fuer Sicherheitsinitialisierungsdatei erstellen.
        if ((md5 = MD5File(szIniFile)) == NULL) {

            // Datei kann nicht uebernommen werden.
            // Eigenen Fehlertext setzen.
            setErrorText( "Erstellen des MD5(2) schlug fehl." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label;
        }
        // Aufgenommene Initialisierungsdatei loeschen
        ::remove(szIniFile);

        sprintf(szFile, "%s.ini", md5);
        // Speicher freigeben.
        free(md5);
        md5 = NULL;

        sprintf(szIniFile, "%s\\%s", path, szFile);
        ::remove(szIniFile); // Vorhandene Initialisierungsdatei loeschen

        // Archivdaten eintragen
        m_iniFile.WriteInt("Archive", "FileDeviceVersion", sizeof(SDFFileDeviceHeader));
        m_iniFile.WriteString("Archive", "Timestamp", timestamp.data());
        if ( versions.size() > 1 )
        {
            int idx = 1;
            std::string key;
            versions.sort( std::greater<std::string>() ); // Liste der Versionen aufsteigend sortieren
            for ( std::string &ver : versions )
            {
                version.clear();
                key = boost::str( boost::format( "Version%1%" ) % idx++ );
                std::copy_n( ver.begin(), 11, std::back_inserter( version ) );

                m_iniFile.WriteString( "Archive", key.data(), version.data() );
            }
        }
        else
        {
            m_iniFile.WriteString("Archive", "Version", versions.begin()->data() );
        }

        m_iniFile.WriteInt("Archive", "Count", fileCount);

        // Initialisierungsdatei schreiben
        m_iniFile.WriteIniFile(szIniFile);
        m_iniFile.CloseIniFile();

        // Datei zu Archiv hinzufuegen
        if (ZipAdd(hZip, szFile, szIniFile) != ZR_OK) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }
        // Aufgenommene Initialisierungsdatei loeschen
        ::remove(szIniFile);

        // Erstellen erfolgreich
        loaded = 1;
        // Kommentar des Archives setzen
        comment = boost::str( boost::format( "Company:\tDatafox GmbH\nContent:\tFiles for MasterIV - devices.\nVersion:\t%1%\nTime stamp:\t%2%") % version % timestamp );
        ZipSetComment(hZip, comment.data());
error_label:
        _findclose(fp);
        // Zipdatei schliessen
        CloseZip(hZip);
        if (md5) free(md5);
        m_iniFile.CloseIniFile();
    } else {

        // Eigenen Fehlertext setzen.
        setErrorText( boost::str( boost::format( "Keine Hex-Dateien im Pfad [ %1% ] gefunden." ) %  path ) );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 

    }

    // Reset der Daten durchfuehren
    resetData();
#endif

    return loaded;
}

int FileDeviceZip::SearchInfosDir(std::string &filePath, const StructHardwareInfo &hwInfo)
{
    int loaded = 0;
    char szFile[260];
#if defined(_MSC_VER) || defined(__MINGW32__)
    char szSearchFile[260];
    intptr_t fp;
    struct _finddata_t finddata;

    sprintf(szSearchFile, "%s\\*.hex", filePath.c_str());
    // Passende Dateien fuer Muster?
    if ((fp = _findfirst(szSearchFile, &finddata)) != -1) {

        // Alle Dateien durchlaufen.
        do {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s\\%s", filePath.c_str(), finddata.name);
#else
    DIR *d;
    struct stat st;
    struct dirent *de;

    if ((d = opendir(filePath.c_str())) != NULL) {

        // Alle Dateien durchlaufen
        while((de = readdir(d)) != NULL) {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s/%s", filePath.c_str(), de->d_name);

            if (stat(szFile, &st) == -1) {
                break;
            }

            if (S_ISREG(st.st_mode)) {

                // Nur ".hex" gueltig.
                int pos = strlen(de->d_name) - 4;
                if (de->d_name[pos] != '.' || 
                    !(de->d_name[pos + 1] == 'h' || de->d_name[pos + 1] == 'H') ||
                    !(de->d_name[pos + 2] == 'e' || de->d_name[pos + 2] == 'E') ||
                    !(de->d_name[pos + 3] == 'x' || de->d_name[pos + 3] == 'X') ) {

                        continue;
                }

#endif	
                // Dateidaten laden
                if (!load(szFile)) {

                    // Umgebung zuruecksetzen
                    resetData();
                    continue;
                }

                // Ist die Softwareinformation in der Geraetedatei verfuegbar?
                if (GetHeaderSoftwareInfo().size) {

                    // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
                    if (CompareInfos(hwInfo, GetHeaderSoftwareInfo())) {

                        loaded = 1;
                        break;
                    }
                } else {
                    continue;
                }
#if defined(_MSC_VER) || defined(__MINGW32__)
            } while (_findnext(fp, &finddata) != -1);
            _findclose(fp);
        }
#else

            }
        } // while

        closedir(d);
    }
#endif

    // Reset bei fehlerhaftem laden
    if (!loaded) {

        resetData();

        // Fehler zuweisen
        setErrorState( Error_state::no_file_match );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }

    return loaded;
}

int FileDeviceZip::CheckZip(const std::string &filePath)
{
    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    char szFile[260], szIniFile[260];
    std::string version;
    char timestamp[20], tmp[256];
    int index, fileDeviceVersion;
    ZIPENTRY ze;
    char *md5 = NULL;

    // Speicher initialisieren
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    memset(timestamp, 0, sizeof(timestamp));
    memset(tmp, 0, sizeof(tmp));

    HZIP hZip = NULL;
    if ((hZip = OpenZip(filePath.c_str(), 0)) == NULL) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szIniFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // MD5 ueber die Initialisierungsdatei erstellen
    if ((md5 = MD5File(szIniFile)) == NULL) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Dateiname zusammenstellen
    sprintf(szFile, "%s.ini", md5);
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile, false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Sicherheitsinitialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szFile)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // Speicher freigeben.
    free(md5);
    md5 = NULL;

    // Informationen der Sicherheitsini laden
    if (!m_iniFile.OpenIniFile(szFile)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei ist gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Version der enthaltenen Geraetedateien pruefen
    fileDeviceVersion = m_iniFile.ReadInt("Archive", "FileDeviceVersion", 0);
    if (static_cast<size_t>( fileDeviceVersion ) > sizeof(SDFFileDeviceHeader)) {

        // Fehlertext zuweisen. Die Software ist nicht aktuell genug fuer diese Geraetedateien.
        setErrorState( Error_state::version );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto error_label;
    }

    // Anzahl der Dateien laden
    //fileCount = m_iniFile.ReadInt("Archive", "Count", 0);

    // Versionen der Dateien laden
    m_version.clear();
    int idx;
    for (idx = 0; idx < 11; ++idx)
    {
        std::string key;
        if (idx == 0)
        {
            key = "Version";
        }
        else
        {
            key = "Version" + std::to_string( idx );
        }

        version = m_iniFile.ReadString("Archive", key.c_str(), "");
        if (idx > 1 && version.size() == 0)
        {
            break;
        }

        if (version.size())
        {
            if (m_version.size())
                m_version += "; ";

            // Version halten
            m_version += version;
        }
    }

    // Zeitstempel des Archives laden
    strcpy(timestamp, m_iniFile.ReadString("Archive", "Timestamp", ""));
    if (!strlen(timestamp)) {
        // Zeitstempel anhand des Erstellungsdatums setzen
        struct stat st;
        if (stat(filePath.c_str(), &st) == 0) {

            struct tm *ptm = localtime(&st.st_ctime);

            sprintf(timestamp, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
                (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
                ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
        }
    }

    // Zeitstempel halten
    m_timestamp = timestamp;
    // Als geladen markieren
    loaded = 1;

error_label:
    // Schliessen der Initialisierungsdatei
    m_iniFile.CloseIniFile();
    // Dateien loeschen
    ::remove(szFile);
    ::remove(szIniFile);
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    // Ziparchiv schliessen
    CloseZip(hZip);	
    if (md5) free(md5);
#endif
    return loaded;
}

int FileDeviceZip::UnzipDefaultDFL(char *fileZIP, char *fileDefaultDFL)
{
    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    char szFile[260], szIniFile[260], szFileDefaultDFL[260];
    char tmp[256];
    int index;
    ZIPENTRY ze;
    char *md5 = NULL;

    // Speicher initialisieren
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    memset(szFileDefaultDFL, 0, sizeof(szFileDefaultDFL));
    memset(tmp, 0, sizeof(tmp));

    resetError();

    HZIP hZip = NULL;
    if ((hZip = OpenZip(fileZIP, 0)) == NULL) {

        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szIniFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // MD5 ueber die Initialisierungsdatei erstellen
    if ((md5 = MD5File(szIniFile)) == NULL) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Dateiname der Sicherheitsini zusammenstellen
    std::sprintf(szFile, "%s.ini", md5);
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile, false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // Speicher freigeben.
    free(md5);
    md5 = NULL;

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    if (!m_iniFile.OpenIniFile(szIniFile)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // MD5 der Textdatedatei pruefen
    // Version der Dateien laden
    strcpy(tmp, m_iniFile.ReadString("Default.dfl", "MD5", ""));
    if (!strlen(tmp)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Ermittlung des MD5 zur Default.dfl fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Suchen der Textdatendatei Default.dfl
    if (ZR_OK != FindZipItem(hZip, "Default.dfl", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szFileDefaultDFL, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Textdatendatei
    if (ZR_OK != UnzipItem(hZip, index, szFileDefaultDFL)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // MD5 ueber die Textdatendatei erstellen
    if ((md5 = MD5File(szFileDefaultDFL)) == NULL) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    if (strcmp(tmp, md5) != 0) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Pruefung des MD5 der Default.dfl gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Kopie der Datei an gewuenschte lokation vornehmen
    if (ZR_OK != UnzipItem(hZip, index, fileDefaultDFL)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Speicher freigeben.
    free(md5);
    md5 = NULL;

error_label:
    // Inidatei schliessen
    m_iniFile.CloseIniFile();
    // Dateien loeschen
    ::remove(szIniFile);
    ::remove(szFileDefaultDFL);
    // Ziparchiv schliessen
    CloseZip(hZip);	
    if (md5) free(md5);
#endif
    return loaded;
}

int FileDeviceZip::SearchInfosZip(std::string &filePath, const StructHardwareInfo &hwInfo)
{
    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    char szFile[260], szIniFile[260];
    std::string version;
    char timestamp[20], tmp[256];
    std::string errorExtended;
    int i, index, fileDeviceVersion;
    ZIPENTRY zeZIP, ze;
    char *md5 = NULL;

    // Speicher initialisieren
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    memset(timestamp, 0, sizeof(timestamp));
    memset(tmp, 0, sizeof(tmp));

    // Fehler zuruecksetzen
    resetError();

    HZIP hZip = NULL;
    if ((hZip = OpenZip(filePath.c_str(), 0)) == NULL) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );

        // Fehler anfuegen
        filePath += ":[" + errorText() + "]";
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szIniFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // MD5 ueber die Initialisierungsdatei erstellen
    if ((md5 = MD5File(szIniFile)) == NULL) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Dateiname zusammenstellen
    sprintf(szFile, "%s.ini", md5);
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile, false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    std::strcpy( szFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
    // Entzippen der Sicherheitsinitialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szFile)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // Speicher freigeben.
    free(md5);
    md5 = NULL;

    // Informationen der Sicherheitsini laden
    if (!m_iniFile.OpenIniFile(szFile)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Version der enthaltenen Geraetedateien pruefen
    fileDeviceVersion = m_iniFile.ReadInt("Archive", "FileDeviceVersion", 0);
    if (static_cast<unsigned>( fileDeviceVersion ) > sizeof(SDFFileDeviceHeader)) {

        // Fehlertext zuweisen. Die Software ist nicht aktuell genug fuer diese Geraetedateien.
        setErrorState( Error_state::version );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto error_label;
    }

    // Anzahl der Dateien laden
    //fileCount = m_iniFile.ReadInt("Archive", "Count", 0);
    // Version der Dateien laden
    version = m_iniFile.ReadString("Archive", "Version", "");
    // Zeitstempel des Archives laden
    strcpy(timestamp, m_iniFile.ReadString("Archive", "Timestamp", ""));
    if (!strlen(timestamp)) {
        // Zeitstempel anhand des Erstellungsdatums setzen
        struct stat st;
        if (stat(filePath.c_str(), &st) == 0) {

            struct tm *ptm = localtime(&st.st_ctime);

            sprintf(timestamp, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
                (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
                ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
        }
    }
    // Schliessen der Initialisierungsdatei
    m_iniFile.CloseIniFile();
    // Datei loeschen
    ::remove(szFile);
    memset(szFile, 0, sizeof(szFile));

    // Informationen zum Archiv ermitteln
    if (ZR_OK != GetZipItem(hZip, -1, &zeZIP)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    if (!m_iniFile.OpenIniFile(szIniFile)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Passende Dateien suchen
    for (i=0; i < zeZIP.index; i++) {

        if (ZR_OK != GetZipItem(hZip, i, &ze)) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            break;
        }

        // Ist kein Ordner?
        if (ze.attr & FILE_ATTRIBUTE_DIRECTORY) {

            continue;
        }
        // Nur ".hex" gueltig.
        size_t pos = strlen(ze.name) - 4;
        if (ze.name[pos] != '.' || 
            !(ze.name[pos + 1] == 'h' || ze.name[pos + 1] == 'H') ||
            !(ze.name[pos + 2] == 'e' || ze.name[pos + 2] == 'E') ||
            !(ze.name[pos + 3] == 'x' || ze.name[pos + 3] == 'X') ) {

                continue;
        }

        // Laden der Softwareinfo
        strcpy(tmp, m_iniFile.ReadString(ze.name, "SWI", ""));
        if (!strlen(tmp) || tmp[0] != '#') {
            // Eigenen Fehlertext setzen.
            setErrorText( "Lesen der Softwareinfo(1) fehlgeschlagen." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label;
        }

        // Die Softwareinfo in eine Struktur wandeln
        StructSoftwareInfo swInfo = SoftwareInfoStringToStruct(tmp);
        if (!swInfo.size) {
            // Eigenen Fehlertext setzen.
            setErrorText( "Lesen der Softwareinfo(2) fehlgeschlagen." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label;
        }

        // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
        if (!CompareInfos(hwInfo, swInfo)) {

            errorExtended.append( boost::str( boost::format( "%1%:[%2%]\n") % ze.name % _deviceInfo.m_errorText ) );

            // Naechste Datei pruefen
            continue;
        }

        // Dateiname zusammenstellen
        std::strcpy( szFile, std::unique_ptr<char[]>( tempnam( nullptr, "df_" ) ).get() );
        // Entzippen der einzuspielenden Datei
        if (ZR_OK != UnzipItem(hZip, i, szFile)) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }
        // MD5 ueber die Initialisierungsdatei erstellen
        if ((md5 = MD5File(szFile)) == NULL) { 
            // Eigenen Fehlertext setzen.
            setErrorText( "Erzeugen des MD5(2) fehlgeschlagen." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label; 
        }

        // MD5 der Datei aus Ini lesen
        strcpy(tmp, m_iniFile.ReadString(ze.name, "MD5", ""));
        if (!strlen(tmp)) {
            // Keine Angabe vorhanden.
            continue;
        }
        if (strcmp(tmp, md5) != 0) {
            // Datei ist korrupt.
            continue;
        }
        // Speicher freigeben.
        free(md5);
        md5 = NULL;

        // Dateidaten laden
        resetData();
        if (!load(szFile)) {
            // Eigenen Fehlertext setzen.
            setErrorText( boost::str( boost::format( "Laden der Geraetedatei fehlgeschlagen (%1%).") % static_cast<int>( errorState() ) ) );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label;
        } else if (GetHeaderSoftwareInfo().size) {

            // Fehler zuruecksetzen
            resetError();

            // Den Namen der gewaehlten Datei fuer Log anfuegen
            filePath += "\\" + std::string(ze.name);
            // Version halten
            m_version = version;
            // Zeitstempel halten
            m_timestamp = timestamp;
            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            loaded = 1;
            break;
        } 
    }

error_label:
    // Ziparchiv schliessen
    CloseZip(hZip);	
    if (md5) free(md5);
    // Inidatei schliessen
    m_iniFile.CloseIniFile();
    // Entzippte Dateien loeschen
    if (strlen(szIniFile)) ::remove(szIniFile);
    if (strlen(szFile)) ::remove(szFile);

    // Reset bei fehlerhaftem laden
    if (errorState() == Error_state::none && !loaded) {

        resetData();

        // Es wurde keine passende Datei gefunden.
        setErrorState( Error_state::no_file_match );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

        // Fehlertext der einzelnen Dateien uebernehmen.
        setErrorTextExtended( errorExtended );
    }
    else 
    { 
        // Fehlertext uebernehmen
        filePath += ":[" + errorText() + "]";
    }
#endif

    return loaded;
}

